# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: qrdm.proto
# Protobuf Python Version: 4.25.2
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\nqrdm.proto\x12\x04qrdm"\xaa\x01\n\tQRContent\x12$\n\x04meta\x18\x01 \x01(\x0b\x32\x16.qrdm.QRContent.QRMeta\x12\x14\n\x0c\x64oc_fragment\x18\x02 \x01(\x0c\x1a\x61\n\x06QRMeta\x12\x15\n\rdocument_hash\x18\x01 \x01(\x06\x12\x17\n\x0fsequence_number\x18\x02 \x01(\r\x12\x16\n\x0etotal_qr_codes\x18\x03 \x01(\r\x12\x0f\n\x07num_ecc\x18\x04 \x01(\r"\x84\x01\n\x0f\x44ocumentPayload\x12\x0f\n\x07\x63ontent\x18\x01 \x01(\x0c\x12\x10\n\x08metadata\x18\x02 \x01(\t\x12\x31\n\tdata_type\x18\x03 \x01(\x0e\x32\x1e.qrdm.DocumentPayload.DataType"\x1b\n\x08\x44\x61taType\x12\x0f\n\x0bUTF8_STRING\x10\x00\x62\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "qrdm_pb2", _globals)
if _descriptor._USE_C_DESCRIPTORS is False:
    DESCRIPTOR._options = None
    _globals["_QRCONTENT"]._serialized_start = 21
    _globals["_QRCONTENT"]._serialized_end = 191
    _globals["_QRCONTENT_QRMETA"]._serialized_start = 94
    _globals["_QRCONTENT_QRMETA"]._serialized_end = 191
    _globals["_DOCUMENTPAYLOAD"]._serialized_start = 194
    _globals["_DOCUMENTPAYLOAD"]._serialized_end = 326
    _globals["_DOCUMENTPAYLOAD_DATATYPE"]._serialized_start = 299
    _globals["_DOCUMENTPAYLOAD_DATATYPE"]._serialized_end = 326
# @@protoc_insertion_point(module_scope)



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Render QRDM PDFs from a set of QR codes, via `reportlab`."""
from __future__ import annotations

import io
import logging
import time
from importlib import resources
from pathlib import Path
from typing import Optional

from qrcode.main import QRCode
from reportlab.graphics import renderPDF
from reportlab.lib.pagesizes import letter
from reportlab.lib.units import inch
from reportlab.lib.utils import ImageReader
from reportlab.pdfgen import canvas
from svglib.svglib import Drawing, svg2rlg

from qrdm import __version__ as QRDM_VERSION

__all__ = ["generate_pdf_pages"]
logger = logging.getLogger(__name__)

# Document Constants
AUTHOR_NAME = f"QRDM v{QRDM_VERSION}"
LOGO_PATH = resources.files("qrdm.qr.data").joinpath("qrdm_logo_red.png")

# Layout Parameters
# Upper left corner of QR region
START_X_PX: float = 0.25 * inch
START_Y_PX: float = 10.25 * inch
# Maximum pixels from the left side of a standard 8.5 inch page
# Make it horizontally symmetric
MAX_X_PX: float = (8.5 * inch) - START_X_PX
# Lower boundary of the QR area, under which the plaintext caption resides
MIN_Y_PX: float = 4.75 * inch
# Spacing of QR codes
QR_MARGIN_PX: float = 0.25 * inch
# Text caption under QR code area
CAPTION_CHAR_WIDTH: int = 192
MAX_CHAR_LIMIT: int = 45 * 192  # num_lines * CAPTION_CHAR_WIDTH


def generate_pdf_pages(
    qr_codes: list[QRCode],
    *,
    buf: io.BufferedIOBase,
    header_text: str = "",
    qr_text: Optional[str] = None,
    filename: Optional[str] = None,
) -> None:
    """Write a QRDM PDF from a list of QRCodes to a binary stream.

    Parameters
    ----------
    qr_codes: list[QRCode]
        List of `qrcode.QRCode` objects, configured to use the `SvgPathImage` image
        factory.
    buf: Writeable binary stream object
        Stream that PDF file will be written to.
    header_text: str, optional
        Text to include as header & footer of QR PDF. Defaults to `""`.
    qr_text: str, optional
        If not `None`, will be used to provide a caption split across the lower portions
        of the generated pages. Defaults to `None`.
    filename: str
        The name of source file that was QR encoded, used to caption the page footer if
        not `None`. Defaults to `None`.

    """
    # Render QR codes as images, and convert to reportlab-compatible Drawings
    qr_imgs = _generate_qr_code_svgs(qr_codes=qr_codes)
    reportlab_qrs = _format_reportlab_img(imgs=qr_imgs)

    # Calculate positions of QR codes based on their dimensions, and whether space
    # is reserved for text content
    include_text = qr_text is not None
    page_qrs, page_qr_positions = _get_qr_positions_per_page(
        reportlab_qrs=reportlab_qrs, include_text=include_text
    )
    # Split up text across number of pages
    if include_text:
        page_qr_text = _split_text_across_pages(qr_text)
    else:
        page_qr_text = []

    logger.debug("QRS_PER_PAGE: %r", page_qrs)
    logger.debug("QRS_POSITIONS: %r", page_qr_positions)

    # Add Sourcing Information
    render_time = time.strftime("%Y-%m-%d %H:%M:%S UTC", time.gmtime())
    if filename is not None:
        file_path = Path(filename)
        printable_filename = file_path.name
        if len(printable_filename) > 40:
            printable_filename = "".join(
                [file_path.stem[:30], " ... .", file_path.suffix]
            )
        doc_title = f"QR Encoding of {filename}"
        footer_text = f"Content from {printable_filename} at {render_time}"
    else:
        printable_filename = None
        doc_title = "QR Encoded Document"
        footer_text = f"Encoded at {render_time}"

    c = canvas.Canvas(buf, pagesize=letter)
    c.setAuthor(AUTHOR_NAME)
    c.setTitle(doc_title)

    for page_index, qr_range in enumerate(page_qrs, start=1):
        logger.debug(f"Redering PDF page {page_index} of {len(page_qrs)}")
        page_footer_text = footer_text + f", Page {page_index} of {len(page_qrs)}"
        _draw_header_footer(c, header_text=header_text, footer_text=page_footer_text)

        for j in range(qr_range[0], qr_range[1]):
            _draw_qr_on_canvas(
                c, rendered_qr=reportlab_qrs[j], pos=page_qr_positions[j]
            )

        if include_text and page_index <= (len(page_qr_text)):
            # Check if we're on the last page, and there's more text chunks than pages
            if (page_index == len(page_qrs)) and (len(page_qr_text) > len(page_qrs)):
                remaining_chars = sum(len(text) for text in page_qr_text[page_index:])
                _add_overflow_notice(c, remaining_chars=remaining_chars)
            _add_page_caption(c, text=page_qr_text[page_index - 1])

        # Cycle to next page
        c.showPage()
    c.save()


def _generate_qr_code_svgs(qr_codes: list[QRCode]) -> list[io.BytesIO]:
    # Sort QR codes by size, so layout algorithm can safely assume that rows won't
    # grow in height
    qr_sizes = [code.version for code in qr_codes]
    # This is just a python list version of `np.argsort`
    code_size_order = sorted(
        range(len(qr_sizes)), key=qr_sizes.__getitem__, reverse=True
    )
    sorted_qr_codes = [qr_codes[ii] for ii in code_size_order]
    imgs = []
    for code in sorted_qr_codes:
        img = code.make_image()
        img_byte_stream = io.BytesIO(initial_bytes=img.to_string())
        imgs.append(img_byte_stream)
    return imgs


def _draw_qr_on_canvas(
    c: canvas.Canvas, *, rendered_qr: Drawing, pos: tuple[int, int]
) -> None:
    # Load the image into reportlab image class, scale and draw
    renderPDF.draw(rendered_qr, c, pos[0], pos[1] - rendered_qr.height)


def _format_reportlab_img(imgs: list[io.BytesIO]) -> list[Drawing]:
    reportlab_qrs = []
    for img in imgs:
        reportlab_qrs.append(svg2rlg(img))
    return reportlab_qrs


def _get_qr_positions_per_page(
    *, reportlab_qrs: list[Drawing], include_text: bool = True
):
    logger.debug("Calculating QR page positions")
    page_qr_positions = []
    page_qrs = []

    if include_text:
        min_y_px = MIN_Y_PX
    else:
        # Maximum pixels from the top of a standard 11 inch page, it
        # includes a .5 inch bottom margin, but no space for text.
        # It does not include .5 top margin, that's included e
        min_y_px = 0.75 * inch

    cur_x_px = START_X_PX
    cur_y_px = START_Y_PX
    last_h_px = 0
    qr_counter = 0

    for ii, rendered_qr in enumerate(reportlab_qrs):
        qr_h_px = rendered_qr.height
        qr_w_px = rendered_qr.width

        enough_horizontal_space = cur_x_px + qr_w_px <= MAX_X_PX
        if not enough_horizontal_space:
            # Start new row, typewriter style
            cur_x_px = START_X_PX
            cur_y_px -= last_h_px + QR_MARGIN_PX
            enough_vertical_space = cur_y_px - qr_h_px >= min_y_px
            if not enough_vertical_space:
                # Start new page
                cur_y_px = START_Y_PX
                page_qrs.append((qr_counter, ii))
                qr_counter = ii

        # Recalculate, as our reference point may have moved
        enough_horizontal_space = cur_x_px + qr_w_px <= MAX_X_PX
        enough_vertical_space = cur_y_px - qr_h_px >= min_y_px
        if enough_horizontal_space and enough_vertical_space:
            # Record the position, and move reference point for next code
            page_qr_positions.append((cur_x_px, cur_y_px))
            cur_x_px += qr_w_px + QR_MARGIN_PX
            last_h_px = qr_h_px
        else:
            raise RuntimeError("Unable to fit QR code on page!")

    page_qrs.append((qr_counter, len(reportlab_qrs)))
    return page_qrs, page_qr_positions


def _split_text_across_pages(text: str) -> list[str]:
    # Make escape characters printable with `repr`
    # Slice [1:] to remove opening quote char, trailing quote handled by the -1 in
    # `breakpoints` below
    caption_text = repr(text)[1:]
    # Chunk text into substings of length MAX_CHAR_LIMIT
    breakpoints = [*list(range(0, len(caption_text), MAX_CHAR_LIMIT)), -1]
    page_text = [
        caption_text[breakpoints[ii] : breakpoints[ii + 1]]
        for ii in range(len(breakpoints) - 1)
    ]
    return page_text


def _draw_header_footer(
    c: canvas.Canvas, *, header_text: str, footer_text: str
) -> None:
    c.drawCentredString(4.25 * inch, 0.25 * inch, header_text)
    c.drawCentredString(4.25 * inch, 10.5 * inch, header_text)
    c.drawCentredString(4.25 * inch, 0.5 * inch, footer_text)
    # Draw QRDM logo
    with LOGO_PATH.open("rb") as imgfile:
        logo_img = ImageReader(imgfile)
        c.drawImage(
            logo_img,
            x=START_X_PX,
            y=START_Y_PX + inch / 4,
            width=inch / 4,
            preserveAspectRatio=True,
            mask="auto",
            anchor="sw",
        )


def _add_page_caption(c: canvas.Canvas, text: str) -> None:
    textobject = c.beginText()
    textobject.setFont("Courier", size=5)
    textobject.setTextOrigin(0.25 * inch, MIN_Y_PX)
    char_line_width = CAPTION_CHAR_WIDTH
    for ii in range(0, len(text), char_line_width):
        textobject.textLine(text[ii : ii + char_line_width])
    c.drawText(textobject)


def _add_overflow_notice(c: canvas.Canvas, *, remaining_chars: int) -> None:
    text = (
        "NOTICE: Remaining source content text omitted due to length. "
        f" ({remaining_chars} characters)"
    )
    c.drawCentredString(4.25 * inch, 0.75 * inch, text)



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Data structures for representing and serializing QR & document payload data."""
from __future__ import annotations

import base64
import json
import zlib
from enum import Enum

from pydantic import BaseModel, ConfigDict, Field, JsonValue

from qrdm.qr import protobuf

__all__ = ["ECSettingValues", "QRMeta", "QRContent", "DocumentPayload"]


class ECSettingValues(str, Enum):
    """Valid settings for QR error correction levels."""

    L = "LOW"
    M = "MEDIUM"
    Q = "QUARTILE"
    H = "HIGH"


class QRMeta(BaseModel):
    """Metadata of individual QRDM QR codes, used for document reconstruction."""

    document_hash: int = Field(ge=0)
    sequence_number: int = Field(ge=0)
    total_qr_codes: int = Field(gt=0)
    num_ecc: int = Field(ge=0)
    model_config = ConfigDict(frozen=True)

    @classmethod
    def _from_proto(cls, pb: protobuf.QRContent.QRMeta) -> QRMeta:
        return cls(
            document_hash=pb.document_hash,
            sequence_number=pb.sequence_number,
            total_qr_codes=pb.total_qr_codes,
            num_ecc=pb.num_ecc,
        )

    def _to_proto(self) -> protobuf.QRContent.QRMeta:
        return protobuf.QRContent.QRMeta(**self.model_dump())


class QRContent(BaseModel):
    """Data content of individual QRDM QR codes."""

    meta: QRMeta
    doc_fragment: bytes = b""

    @classmethod
    def model_validate_protobuf_bytes(cls, b: bytes) -> QRContent:
        """Parse serialized binary data into a `QRContent`."""
        pb = protobuf.QRContent()
        pb.ParseFromString(b)
        meta = QRMeta._from_proto(pb.meta)
        return cls(meta=meta, doc_fragment=pb.doc_fragment)

    @classmethod
    def model_validate_b85_bytes(cls, b: bytes) -> QRContent:
        """Parse b85-encodeded & serialized binary data into a `QRContent`."""
        decoded_bytes = base64.b85decode(b)
        return cls.model_validate_protobuf_bytes(decoded_bytes)

    def _to_proto(self) -> protobuf.QRContent:
        meta_pb = self.meta._to_proto()
        return protobuf.QRContent(meta=meta_pb, doc_fragment=self.doc_fragment)

    def model_dump_protobuf_bytes(self) -> bytes:
        """Serialize this object into binary data."""
        pb = self._to_proto()
        return pb.SerializeToString()

    def model_dump_b85_bytes(self) -> bytes:
        """Serialize this object into b85-encoded binary data."""
        raw_bytes = self.model_dump_protobuf_bytes()
        return base64.b85encode(raw_bytes)


class DocumentPayload(BaseModel):
    """Represents and serializes QRDM source documents."""

    content: str = ""
    metadata: JsonValue = None
    model_config = ConfigDict(frozen=True)

    @classmethod
    def model_validate_protobuf_bytes(cls, b: bytes) -> DocumentPayload:
        """Parse serialized binary data into a `DocumentPayload`."""
        pb = protobuf.DocumentPayload()
        pb.ParseFromString(b)
        meta_json = None
        if len(pb.metadata) > 0:
            meta_json = json.loads(pb.metadata)
        if pb.data_type != protobuf.DocumentPayload.DataType.UTF8_STRING:
            raise ValueError("Unknown data type")
        return cls(content=pb.content.decode("utf-8"), metadata=meta_json)

    @classmethod
    def model_validate_compressed_bytes(cls, b: bytes) -> DocumentPayload:
        """Parse serialized and compressed binary data into a `DocumentPayload`."""
        inflated_bytes = zlib.decompress(b)
        return cls.model_validate_protobuf_bytes(inflated_bytes)

    def _to_proto(self) -> protobuf.DocumentPayload:
        meta_str = None
        if self.metadata is not None:
            meta_str = json.dumps(self.metadata)
        return protobuf.DocumentPayload(
            content=self.content.encode("utf-8"),
            metadata=meta_str,
            data_type=protobuf.DocumentPayload.DataType.UTF8_STRING,
        )

    def model_dump_protobuf_bytes(self) -> bytes:
        """Serialize this object into binary data."""
        pb = self._to_proto()
        return pb.SerializeToString()

    def model_dump_compressed_bytes(self) -> bytes:
        """Serialize this object into compressed binary data."""
        raw_bytes = self.model_dump_protobuf_bytes()
        return zlib.compress(raw_bytes, level=zlib.Z_BEST_COMPRESSION)



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""QRDM provides tools for encoding and decoding documents as a series of QR codes."""
__version__ = "2.0.0"

from qrdm.qr.decode import decode_qr_pdf
from qrdm.qr.encode import encode_qr_pdf

__all__ = ["decode_qr_pdf", "encode_qr_pdf"]



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""QRDM-specific exception classes."""

__all__ = ["QRDecodeError", "QREncodeError"]


class QRDecodeError(Exception):
    """Unable to reconstruct a source document from an input file."""

    pass


class QREncodeError(Exception):
    """Unable to construct a QR PDF from an input source document."""

    pass



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""FastAPI application router for QR encode & decode functions."""
from __future__ import annotations

import io
from pathlib import Path
from typing import Annotated, Optional

import structlog
from fastapi import APIRouter, Depends, File, Form, HTTPException, UploadFile
from fastapi.responses import Response
from pydantic import BaseModel, Json, JsonValue

from qrdm.backend.config import BackendSettings, get_backend_settings
from qrdm.exceptions import QRDecodeError, QREncodeError
from qrdm.models import DocumentPayload
from qrdm.qr.decode import decode_qr_pdf
from qrdm.qr.encode import encode_qr_pdf, get_file_content

__all__ = ["router"]
logger = structlog.get_logger()

router = APIRouter(tags=["QR Encoding & Decoding"])


class EncodeData(BaseModel):
    """`/encode` endpoint request data."""

    header_text: str = ""
    encode_ec_codes: Optional[bool] = None
    metadata: JsonValue = None
    encoding: Optional[str] = None

    @classmethod
    def as_form_data(
        cls,
        header_text: Annotated[str, Form()] = "",
        metadata: Annotated[Json, Form()] = None,
        encoding: Annotated[Optional[str], Form()] = None,
    ) -> EncodeData:
        """Wrap the`EncodeData` constructor to accept HTTP Form data."""
        return cls(header_text=header_text, metadata=metadata, encoding=encoding)


@router.post("/encode", summary="Encode a plaintext document as a PDF of QR codes")
def encode_qr_endpoint(
    document: Annotated[UploadFile, File()],
    settings: Annotated[BackendSettings, Depends(get_backend_settings)],
    body: EncodeData = Depends(EncodeData.as_form_data),
) -> Response:
    # FastAPI uses Markdown formatting for docstring rendering in Swagger, etc.
    """Convert a plaintext document to a PDF containing the document encoded as QR codes.

    Required Form Data
    ------------------
    - **document** (binary): File to be QR-encoded
    - **header_text** (string, optional): Text to include as header & footer of QR PDF
    - **metadata** (string, optional): Text in this field will be treated as JSON, and
      encoded alongside the document contents in the QR PDF.
    - **encoding** (str, optional): Encoding used to decode the document file data into
      a string. If not provided, defaults to automatic detection via the
      `charset_normalizer` package. Valid encodings can be found in the Python
      documentation: https://docs.python.org/3.9/library/codecs.html#standard-encodings
    """
    # upload_file = body.upload_file
    file_data = document.file.read()
    if document.filename is not None:
        upload_filename = Path(document.filename)
    else:
        upload_filename = Path("document")
    structlog.contextvars.bind_contextvars(
        action="QR Encode", input_file={"filename": str(upload_filename)}
    )

    try:
        file_content = get_file_content(file_data, encoding=body.encoding)
        logger.debug("Extracted file content: %r", file_content)
        output_file_data = encode_qr_pdf(
            document_content=file_content,
            header_text=body.header_text,
            metadata=body.metadata,
            document_name=upload_filename.name,
            encode_ec_codes=settings.QRDM_ENCODE_EC_CODES,
            error_tolerance=settings.QRDM_ERROR_TOLERANCE,
        )
    except Exception as e:
        logger.exception(e, success=False)
        if isinstance(e, QREncodeError):
            msg = str(e)
        else:
            msg = "ERROR: Unable to convert file to QR PDF"
        headers = {"QRErrorType": e.__class__.__name__}
        raise HTTPException(status_code=500, detail=msg, headers=headers)

    output_filename = upload_filename.stem + "_qr.pdf"
    logger.info(
        f"Completed QR Conversion of {upload_filename!s}",
        output_file={"filename": output_filename},
    )
    return Response(
        content=output_file_data,
        media_type="application/pdf",
        headers={"Content-Disposition": f'attachment; filename="{output_filename}"'},
    )


@router.post(
    "/decode",
    response_model_exclude_none=True,
    summary="Reconstruct a source document from a QRDM PDF",
)
def decode_qr_endpoint(document: Annotated[UploadFile, File()]) -> DocumentPayload:
    # FastAPI uses Markdown formatting for docstring rendering in Swagger, etc.
    """Reconstruct a source document from a QRDM PDF.

    Required Form Data
    ------------------
    - **document** (binary): QRDM PDF file to be decoded

    Returns
    -------
    JSON with fields:
    - **content** (string): Original document contents
    - **metadata** (optional): JSON representation of data provided to the `metadata`
      field of the `/encode` endpoint.

    """
    file_data = io.BytesIO(document.file.read())
    structlog.contextvars.bind_contextvars(
        action="QR Decode", input_file={"filename": str(document.filename)}
    )

    try:
        decoded_content = decode_qr_pdf(file_data)
        if decoded_content is None:
            raise QRDecodeError("Could not locate any QR codes in document.")
    except Exception as e:
        logger.exception(e, success=False)
        if isinstance(e, QRDecodeError):
            msg = str(e)
        else:
            msg = "ERROR: Unable to decode QR File"
        headers = {"QRErrorType": e.__class__.__name__}
        raise HTTPException(status_code=500, detail=msg, headers=headers)
    logger.info(f"Completed QR Decode of {document.filename!s}")
    return decoded_content



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Recover source documents from QRDM PDFs."""
from __future__ import annotations

import concurrent.futures
import hashlib
import io
import logging
import multiprocessing
import time
import warnings
from itertools import chain
from pathlib import Path
from typing import Optional, Union

import fitz
import reedsolo
import structlog.contextvars
from PIL import Image, ImageFilter
from pyzbar.pyzbar import Decoded, ZBarSymbol, decode

from qrdm.exceptions import QRDecodeError
from qrdm.models import DocumentPayload, QRContent

__all__ = ["decode_qr_pdf"]
logger = logging.getLogger(__name__)


def decode_qr_pdf(
    input_file: Union[str, Path, bytes, io.BytesIO]
) -> Optional[DocumentPayload]:
    """Decode a QRDM PDF into the original document content and metadata.

    Parameters
    ----------
    input_file: Path-like, bytes, or io.BytesIO
        Path to PDF file, binary stream, or bytes containing the file contents of the QR
        PDF. (e.g. as returned by `encode_qr_pdf`) Passed as input to the `fitz.open`
        function provided by PyMuPDF.

    Returns
    -------
    decoded_document: DocumentPayload or `None`
        If no QR codes are detected in the PDF by pyzbar, `None` is returned. Otherwise,
        returns a Pydantic model containing the encoded document & metadata, with
        attributes:

            - content: str - Source document contents
            - metadata: JsonValue, optional - Document metadata, as provided to the
              `metadata` argument of the encoding.

    Examples
    --------
    >>> qrdm.encode_qr_pdf(
    ...     "Lorem ipsum dolor sit amet. 😎",
    ...     metadata={"example": True},
    ...     path="example_qr_pdf.pdf",
    ... )
    >>> qrdm.decode_qr_pdf("example_qr_pdf.pdf")
    DocumentPayload(content='Lorem ipsum dolor sit amet. 😎', metadata={'example': True})

    """
    if isinstance(input_file, (str, Path)):
        doc = fitz.open(input_file, filetype="pdf")
    else:
        if isinstance(input_file, bytes):
            stream = io.BytesIO(input_file)
        else:
            stream = input_file
        doc = fitz.open(stream=stream, filetype="pdf")

    extracted_images = _extract_page_images(doc)
    extracted_contents = _extract_qr_contents_from_images(extracted_images)
    logger.debug(f"Number of extracted payloads: {len(extracted_contents)}")

    if len(extracted_contents) == 0:
        warnings.warn("Failed to extract any QR data from input document")
        return None

    try:
        logger.debug("Reconstructing QR Data")
        decoded_document = _reconstruct_document_payload(
            extracted_contents=extracted_contents
        )
    except Exception as e:
        msg = f"ERROR: Failed to decode PDF, found {len(extracted_contents)} QR codes"
        logger.error(msg)
        raise e

    return decoded_document


def _extract_page_images(doc: fitz.Document) -> list[Image.Image]:
    """Render PDF pages as a sequence of PNG images."""
    logger.debug("Rendering PDF pages")
    extracted_images = []
    for page in doc:
        page_pixmap: fitz.Pixmap = page.get_pixmap(dpi=300)
        extracted_image = Image.open(io.BytesIO(page_pixmap.tobytes("png")))
        extracted_images.append(extracted_image)
    logger.debug("Rendered %d images from PDF Document", len(extracted_images))
    return extracted_images


def _extract_qr_contents_from_images(images: list[Image.Image]) -> dict[int, QRContent]:
    """Decode QRDM QR code payloads contained in a sequence of images.

    Includes attempts at image enhancement if the expected number of QR codes are not
    successfully decoded.
    """
    try:
        max_workers = multiprocessing.cpu_count()
    except NotImplementedError as e:
        logger.warning(f"Failed to get cpu count for multithreading with error: {e}")
        logger.debug("Defaulting to 10 worker threads")
        max_workers = min(10, len(images))

    extracted_contents = {}
    logger.debug(f"Decoding QRs from {len(images)} input images")
    start_time = time.time()
    extracted_contents = _batch_decode_qr_imgs(images, max_workers=max_workers)
    total_time = time.time() - start_time
    logger.debug(f"Decoded {len(extracted_contents)} codes in {total_time:0.4f}s")

    # Check if enough qrs were found for document decoding
    if not _sufficient_decodes(extracted_contents):
        logger.debug("Not enough QRs decoded, attempting image enhancement")
        new_contents = _batch_filter_and_decode_qr_imgs(
            images, existing_contents=extracted_contents, max_workers=max_workers
        )
        extracted_contents.update(new_contents)
        logger.debug(
            "Total QR codes decoded after image enhancement: %d",
            len(extracted_contents),
        )

    qr_info = structlog.contextvars.get_contextvars().get("qr_info", {})
    qr_info.update(
        {
            "extracted_qrs": list(extracted_contents.keys()),
            "num_extracted_qrs": len(extracted_contents),
        }
    )
    structlog.contextvars.bind_contextvars(qr_info=qr_info)

    return extracted_contents


def _reconstruct_document_payload(
    extracted_contents: dict[int, QRContent]
) -> DocumentPayload:
    """Reconstruct a DocumentPayload from a sequence of individual QR payloads."""
    first_content = next(iter(extracted_contents.values()))
    total_qr_codes = first_content.meta.total_qr_codes
    num_ecc = first_content.meta.num_ecc
    min_qrs_needed = total_qr_codes - num_ecc

    if len(extracted_contents) < min_qrs_needed:
        msg = (
            "Insufficient QR payloads for recovery. "
            + f"Need {min_qrs_needed}, got {len(extracted_contents)}"
        )
        raise QRDecodeError(msg)

    doc_fragments: dict[int, bytes] = {
        seq: contents.doc_fragment for (seq, contents) in extracted_contents.items()
    }
    max_payload_len = max(len(fragment) for fragment in doc_fragments.values())

    dropped_code_inds = []
    for seq_num in range(total_qr_codes):
        if seq_num not in extracted_contents:
            logger.debug(f"Missing code: {seq_num}")
            dropped_code_inds.append(seq_num)
            doc_fragments[seq_num] = b"\0" * max_payload_len

    sorted_payload_tuples = sorted(doc_fragments.items(), key=lambda tuple: tuple[0])
    sorted_payloads = [t[1] for t in sorted_payload_tuples]
    logger.debug(
        "Directly extracted payloads: %r", [p.hex(" ", 4) for p in sorted_payloads]
    )

    if num_ecc > 0:
        logger.debug("Checking error-correction codes")
        received_symbols = list(zip(*sorted_payloads))
        decoding_codec = reedsolo.RSCodec(num_ecc)
        fixed_symbols = [
            decoding_codec.decode(x, erase_pos=dropped_code_inds, only_erasures=True)[0]
            for x in received_symbols
        ]
        sorted_payloads = [bytes(x) for x in zip(*fixed_symbols)]
        logger.debug("Post-EC payloads: %r", [p.hex(" ", 4) for p in sorted_payloads])

    total_payload_bytes = b"".join(sorted_payloads).strip(b"\0")
    logger.debug("Concatenated total payload: %s", total_payload_bytes.hex(" ", 4))

    h = hashlib.shake_256(total_payload_bytes)
    hash_bytes_check = h.digest(8)
    document_hash_check = int.from_bytes(hash_bytes_check, "big")
    logger.debug("Recovered document hash: %s", hash_bytes_check.hex(" ", 4))
    if document_hash_check != first_content.meta.document_hash:
        logger.error(
            "Recovered document does not match verification hash: "
            f"{document_hash_check} vs. {first_content.meta.document_hash}"
        )
        raise QRDecodeError("Recovered document does not match checksum")

    output = DocumentPayload.model_validate_compressed_bytes(total_payload_bytes)

    logger.debug("Recomposed document (complete): %r", output)

    return output


def _batch_decode_qr_imgs(
    images: list[Image.Image],
    max_workers: Optional[int] = None,
    threaded_executor: Optional[concurrent.futures.ThreadPoolExecutor] = None,
) -> dict[int, QRContent]:
    """Decode a sequence of QR images into their corresponding data payloads."""
    # Multithreaded Mode
    if threaded_executor:
        # Use provided thread pool
        decodes = list(threaded_executor.map(_pyzbar_decode_qr_job, images))
    else:
        # Initialize threadpool for single use
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            decodes = list(executor.map(_pyzbar_decode_qr_job, images))
    # unpacking results
    flattened_decodes = []
    flattened_decodes = list(chain(*decodes))
    qr_contents = _parse_qr_contents(flattened_decodes)

    qr_info = structlog.contextvars.get_contextvars().get("qr_info", {})
    qr_info.update(
        {
            "qr_meta": next(iter(qr_contents.values())).meta.model_dump(
                exclude={"sequence_number"}
            )
        }
    )
    structlog.contextvars.bind_contextvars(qr_info=qr_info)

    return qr_contents


def _pyzbar_decode_qr_job(img: Image.Image) -> list[Decoded]:
    return decode(img, symbols=[ZBarSymbol.QRCODE])


def _parse_qr_contents(decodes: list[Decoded]) -> dict[int, QRContent]:
    """Parse a list of pyzbar decodes into ordered set of QRContents."""
    extracted_contents = {}
    for ii, qr in enumerate(decodes):
        logger.debug(f"Parsing decode {ii+1} of {len(decodes)}")
        try:
            content = QRContent.model_validate_b85_bytes(qr.data)
        except Exception as e:
            raise QRDecodeError("Could not read QR payload") from e
        extracted_contents[content.meta.sequence_number] = content

        logger.debug(
            "Recovered QR #%d payload with metadata: %r",
            content.meta.sequence_number,
            dict(content.meta),
        )
        logger.debug(
            "Recovered QR #%d payload with content: %s",
            content.meta.sequence_number,
            content.doc_fragment.hex(" ", 4),
        )

    return extracted_contents


def _batch_filter_and_decode_qr_imgs(
    images: list[Image.Image],
    existing_contents: Optional[dict[int, QRContent]] = None,
    max_workers: Optional[int] = None,
) -> dict[int, QRContent]:
    """Apply a image filters to page images to attempt to decode additional QR codes."""
    if existing_contents is None:
        existing_contents = {}
    extracted_contents = existing_contents.copy()
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
        for blur_radius in [2, 3, 4]:
            # blur all images
            blurred_imgs = [_box_blur(img, blur_radius) for img in images]
            # run zbar
            new_contents = _batch_decode_qr_imgs(
                blurred_imgs, threaded_executor=executor
            )
            extracted_contents.update(new_contents)
            # check if sufficient decodes
            if _sufficient_decodes(extracted_contents):
                break

            # double blur all images
            db_blurred_imgs = [_box_blur(img, blur_radius) for img in blurred_imgs]
            # run zbar
            new_contents = _batch_decode_qr_imgs(
                db_blurred_imgs, threaded_executor=executor
            )
            extracted_contents.update(new_contents)
            # check if sufficient decodes
            if _sufficient_decodes(extracted_contents):
                break
    return extracted_contents


def _box_blur(img: Image.Image, kernel_size: int) -> Image.Image:
    """Apply a box filter blur of `kernel_size` to an image."""
    blurred_img = img.filter(ImageFilter.BoxBlur(kernel_size))
    return blurred_img


def _sufficient_decodes(qr_contents: dict[int, QRContent]) -> bool:
    """Check if QR code contents are sufficient to reconstruct the source document."""
    decoded_qr_count = len(qr_contents)
    if decoded_qr_count == 0:
        return False
    one_code = next(iter(qr_contents.values()))
    required_qrs = one_code.meta.total_qr_codes - one_code.meta.num_ecc
    sufficient_decodes = decoded_qr_count >= required_qrs
    if not sufficient_decodes:
        logger.debug("Missing codes: %r", _list_missing_qr_codes(qr_contents))
    return sufficient_decodes


def _list_missing_qr_codes(qr_contents: dict[int, QRContent]) -> list[int]:
    """Determine indices of missing QR code contents."""
    one_code = next(iter(qr_contents.values()))
    total_qrs = one_code.meta.total_qr_codes
    return [ii for ii in range(total_qrs) if ii not in qr_contents]



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
import csv
from enum import IntEnum
from importlib import resources

from qrcode import constants as qrconstants

__all__ = [
    "BOX_SIZE",
    "EC_CODE_PROPORTION",
    "ErrorCorrectionLevel",
    "QR_CAPACITIES",
    "QR_SIZE",
]

# Fits a grid of codes cleanly into US Letter PDF dimensions with pixel size 6
QR_SIZE: int = 22
BOX_SIZE: int = 6
# How many error-correction QR codes to generate per QR code of document content
# Applied as ceil(num_qr * EC_CODE_PROPORTION)
EC_CODE_PROPORTION: float = 0.2


class ErrorCorrectionLevel(IntEnum):
    L = qrconstants.ERROR_CORRECT_L
    M = qrconstants.ERROR_CORRECT_M
    Q = qrconstants.ERROR_CORRECT_Q
    H = qrconstants.ERROR_CORRECT_H


QR_CAPACITIES: dict[int, dict[str, int]] = {}
qr_capacity_file_path = resources.files("qrdm.qr.data").joinpath("qr_capacity.csv")
with qr_capacity_file_path.open() as f:
    reader = csv.DictReader(f)
    for ii, row in enumerate(reader, start=1):
        QR_CAPACITIES[ii] = {
            level.name: int(row[level.name]) for level in ErrorCorrectionLevel
        }



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Construct QRDM FastAPI Application."""

import uuid

import structlog
from fastapi import FastAPI, Request, Response
from uvicorn.protocols.utils import get_path_with_query_string

from qrdm import __version__ as VERSION
from qrdm.backend.routers import qr

__all__ = ["app"]
access_logger = structlog.stdlib.get_logger("qrdm.access")

app = FastAPI(title="QRDM", description="QR Data Manager", version=VERSION)


# Adapted from https://gist.github.com/nymous/f138c7f06062b7c43c060bf03759c29e
# MIT License - Copyright (c) 2023 Thomas GAUDIN
@app.middleware("http")
async def logging_middleware(request: Request, call_next) -> Response:
    """Middleware function for logging all incoming network requests."""
    structlog.contextvars.clear_contextvars()
    request_id = str(uuid.uuid4())
    structlog.contextvars.bind_contextvars(request_id=request_id)
    url = get_path_with_query_string(request.scope)
    if request.client is not None:
        client_host = request.client.host
        client_port = request.client.port
    else:
        client_host, client_port = None, None
    http_method = request.method
    http_version = request.scope["http_version"]

    # Pass request through to destination
    response = await call_next(request)

    # Recreate the Uvicorn access log format
    status_code = response.status_code
    request_log_str = f"{http_method} {url} HTTP/{http_version} - {status_code}"
    if request.client is not None:
        request_log_str = " - ".join([f"{client_host}:{client_port}", request_log_str])
    access_logger.info(
        request_log_str,
        http={
            "url": str(request.url),
            "status_code": status_code,
            "method": http_method,
            "version": http_version,
        },
        network={"client": {"ip": client_host, "port": client_port}},
    )
    return response


@app.get("/version", tags=["Application"], summary="Application version")
def get_version():
    """Return the version of the running QRDM application."""
    resp = {"version": app.version}
    return resp


app.include_router(qr.router)



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
from typing import ClassVar as _ClassVar
from typing import Mapping as _Mapping
from typing import Optional as _Optional
from typing import Union as _Union

from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper

DESCRIPTOR: _descriptor.FileDescriptor

class QRContent(_message.Message):
    __slots__ = ("meta", "doc_fragment")

    class QRMeta(_message.Message):
        __slots__ = ("document_hash", "sequence_number", "total_qr_codes", "num_ecc")
        DOCUMENT_HASH_FIELD_NUMBER: _ClassVar[int]
        SEQUENCE_NUMBER_FIELD_NUMBER: _ClassVar[int]
        TOTAL_QR_CODES_FIELD_NUMBER: _ClassVar[int]
        NUM_ECC_FIELD_NUMBER: _ClassVar[int]
        document_hash: int
        sequence_number: int
        total_qr_codes: int
        num_ecc: int
        def __init__(
            self,
            document_hash: _Optional[int] = ...,
            sequence_number: _Optional[int] = ...,
            total_qr_codes: _Optional[int] = ...,
            num_ecc: _Optional[int] = ...,
        ) -> None: ...

    META_FIELD_NUMBER: _ClassVar[int]
    DOC_FRAGMENT_FIELD_NUMBER: _ClassVar[int]
    meta: QRContent.QRMeta
    doc_fragment: bytes
    def __init__(
        self,
        meta: _Optional[_Union[QRContent.QRMeta, _Mapping]] = ...,
        doc_fragment: _Optional[bytes] = ...,
    ) -> None: ...

class DocumentPayload(_message.Message):
    __slots__ = ("content", "metadata", "data_type")

    class DataType(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
        __slots__ = ()
        UTF8_STRING: _ClassVar[DocumentPayload.DataType]

    UTF8_STRING: DocumentPayload.DataType
    CONTENT_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    DATA_TYPE_FIELD_NUMBER: _ClassVar[int]
    content: bytes
    metadata: str
    data_type: DocumentPayload.DataType
    def __init__(
        self,
        content: _Optional[bytes] = ...,
        metadata: _Optional[str] = ...,
        data_type: _Optional[_Union[DocumentPayload.DataType, str]] = ...,
    ) -> None: ...



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Generate QRDM PDFs from source documents."""
from __future__ import annotations

import hashlib
import io
import logging
import shutil
from collections.abc import Iterator
from math import ceil
from pathlib import Path
from typing import Optional, Union, overload

import qrcode.image.svg as qr_svg
import reedsolo
import structlog
from charset_normalizer import from_bytes
from pydantic import JsonValue
from qrcode.main import QRCode
from structlog import contextvars

import qrdm.qr._const as qr_const
from qrdm.exceptions import QREncodeError
from qrdm.models import DocumentPayload, ECSettingValues, QRContent, QRMeta
from qrdm.qr.pdf_writer import generate_pdf_pages
from qrdm.qr.protobuf import PROTOBUF_RESERVED_LEN

__all__ = ["encode_qr_pdf"]
logger = logging.getLogger(__name__)

# Sequence number recorded as 32-bit uint
N_MAX_QRS = 2**32

PathLike = Union[str, Path]


@overload
def encode_qr_pdf(
    document_content: str,
    *,
    path: None = ...,
    header_text: str = ...,
    metadata: Optional[JsonValue] = ...,
    document_name: Optional[str] = ...,
    encode_ec_codes: bool = ...,
    error_tolerance: Union[str, ECSettingValues] = ...,
) -> bytes: ...


@overload
def encode_qr_pdf(
    document_content: str,
    *,
    path: Union[PathLike, io.BufferedIOBase] = ...,
    header_text: str = ...,
    metadata: Optional[JsonValue] = ...,
    document_name: Optional[str] = ...,
    encode_ec_codes: bool = ...,
    error_tolerance: Union[str, ECSettingValues] = ...,
) -> None: ...


def encode_qr_pdf(
    document_content: str,
    *,
    path: Optional[Union[PathLike, io.BufferedIOBase]] = None,
    header_text: str = "",
    metadata: Optional[JsonValue] = None,
    document_name: Optional[str] = None,
    encode_ec_codes: bool = True,
    error_tolerance: Union[str, ECSettingValues] = ECSettingValues.M,
) -> Union[None, bytes]:
    """Convert a plaintext document to a PDF containing the document encoded as QR codes.

    Parameters
    ----------
    document_content: str
        Document to be QR-encoded, as a python string.
    path: str, path object, file-like object, or `None`, optional
        String, path object (implementing os.PathLike[str]), or file-like object
        implementing a binary write() function. If None, the result is returned as
        bytes. Defaults to `None`.
    header_text: str, optional
        Text to include as header & footer of QR PDF. Defaults to `""`.
    metadata: JsonValue, optional
        JSON-encodable data that is encoded alongside the document contents in the QR
        PDF. (I.e. may be a JSON-scalar `[str|bool|int|float]` or a `dict[str,
        JsonValue]`)
    document_name: str, optional
        Name of the original file or document, to label output PDF.
    encode_ec_codes: bool, optional
        Whether to include additional error-correcting QR codes. These are useful when
        individual QR codes fail to scan.
    error_tolerance: One of `"LOW", "MEDIUM", "QUARTILE", "HIGH"`, optional
        Error correction level of the individual QR codes. Defaults to `"MEDIUM"`.

    Returns
    -------
    File data as bytes if the `path` argument is not provided, otherwise `None`.

    Examples
    --------
    >>> file_data = qrdm.encode_qr_pdf("This returns bytes directly", path=None)
    >>> qrdm.encode_qr_pdf("This is being written directly to a file", path="qr.pdf")
    >>> with open("stream.pdf", "wb") as f:
    ...     qrdm.encode_qr_pdf("This is being written to a file-like stream", path=f)

    """
    buf: io.BufferedIOBase
    if path is None or isinstance(path, (str, Path)):
        buf = io.BytesIO()
    else:
        buf = path
    if isinstance(error_tolerance, str):
        error_tolerance = ECSettingValues(error_tolerance)

    document_payload = DocumentPayload(content=document_content, metadata=metadata)
    qr_payloads = generate_qr_payloads(
        document_payload,
        encode_ec_codes=encode_ec_codes,
        error_tolerance=error_tolerance,
    )
    qr_codes = generate_qr_codes(qr_payloads, error_tolerance=error_tolerance)

    generate_pdf_pages(
        qr_codes,
        buf=buf,
        header_text=header_text,
        qr_text=document_payload.content,
        filename=document_name,
    )
    if path is None or isinstance(path, (str, Path)):
        buf.seek(0)
        if isinstance(path, (str, Path)):
            with open(path, "wb") as f:
                shutil.copyfileobj(buf, f)
        else:
            return buf.read()


def get_file_content(file_data: bytes, encoding: Optional[str] = None) -> str:
    """Extract text file contents from binary file data, with optional encoding detection.

    Parameters
    ----------
    file_data: bytes
        Binary file data}
    encoding: str or `None`, optional
        Encoding used to decode the document file data into a string. If not provided,
        defaults to automatic detection via the `charset_normalizer` package. Valid
        encodings can be found in the Python documentation:
        https://docs.python.org/3.9/library/codecs.html#standard-encodings

    Returns
    -------
    file_content: str
        Decoded file contents

    """
    input_file_info = structlog.contextvars.get_contextvars().get("input_file", {})
    input_file_info["autodetect_encoding"] = encoding is None
    if encoding is None:
        decoded_content = from_bytes(file_data).best()
        if decoded_content is None:
            raise QREncodeError("Could not determine valid encoding for file contents")
        encoding = decoded_content.encoding
        file_content = str(decoded_content)
    else:
        try:
            file_content = file_data.decode(encoding)
        except ValueError as e:
            raise QREncodeError(
                f"Could not decode file with encoding: {encoding}"
            ) from e
    logger.debug("Decoded file contents with: %s", encoding)
    input_file_info["encoding"] = encoding
    structlog.contextvars.bind_contextvars(input_file=input_file_info)
    return file_content


def generate_qr_payloads(
    document_content: DocumentPayload,
    *,
    encode_ec_codes: bool,
    error_tolerance: ECSettingValues = ECSettingValues.M,
) -> list[bytes]:
    """Construct b85-encoded payloads from a document for QR code construction."""
    logger.debug("Generating QR code payloads")
    total_payload_bytes = document_content.model_dump_compressed_bytes()
    logger.debug("Compressed document: %s", total_payload_bytes.hex(" ", 4))

    h = hashlib.shake_256(total_payload_bytes)
    hash_bytes = h.digest(8)
    logger.debug("Document hash: %s", hash_bytes.hex(" ", 4))
    document_hash = int.from_bytes(hash_bytes, "big")

    max_qr_bytes = qr_const.QR_CAPACITIES[qr_const.QR_SIZE][error_tolerance.name]
    single_qr_payload_size = max_qr_bytes - PROTOBUF_RESERVED_LEN
    # Reduce payload size to account for b85 inflation
    single_qr_payload_size = (single_qr_payload_size // 5) * 4

    # If document is small enough to fit in a single code, skip `_split_file_content` to
    # avoid the null-byte padding
    if len(total_payload_bytes) <= single_qr_payload_size:
        payload_fragments = [total_payload_bytes]
    else:
        payload_fragments = list(
            _split_file_content(
                total_payload_bytes, maximum_length=single_qr_payload_size
            )
        )

    if encode_ec_codes:
        num_ecc = ceil(len(payload_fragments) * qr_const.EC_CODE_PROPORTION)
        payload_fragments = _generate_ec_fragments(payload_fragments, num_ecc=num_ecc)
    else:
        num_ecc = 0

    total_qr_codes = len(payload_fragments)
    if total_qr_codes >= N_MAX_QRS:
        msg = (
            f"The total number of QR codes: {total_qr_codes} exceeds the maximum of "
            f"{N_MAX_QRS}, consider breaking the file into smaller pieces."
        )
        raise QREncodeError(msg)

    qr_contents: list[bytes] = []
    for sequence_number, fragment in enumerate(payload_fragments):
        qr_meta = QRMeta(
            document_hash=document_hash,
            sequence_number=sequence_number,
            total_qr_codes=total_qr_codes,
            num_ecc=num_ecc,
        )
        logger.debug(
            "Constructing QR #%d payload with metadata: {%s}",
            sequence_number,
            dict(qr_meta),
        )
        logger.debug(
            "Constructing QR #%d payload with content: %s",
            sequence_number,
            fragment.hex(" ", -4),
        )
        qr_content = QRContent(meta=qr_meta, doc_fragment=fragment)
        qr_b64_content = qr_content.model_dump_b85_bytes()
        logger.debug("QR b85-encoded size: %d", len(qr_b64_content))
        qr_contents.append(qr_b64_content)

        contextvars.bind_contextvars(
            qr_info={
                "encode_ec_codes": encode_ec_codes,
                "qr_meta": qr_meta.model_dump(exclude={"sequence_number"}),
            }
        )

    return qr_contents


def _split_file_content(content: bytes, *, maximum_length: int) -> Iterator[bytes]:
    """Split an array of bytes into chunks of *equal* length n, null padding when required.

    Equal-length chunks are required for generating error-correction QR codes.
    """
    while len(content) > maximum_length:
        yield content[:maximum_length]
        content = content[maximum_length:]
    yield content + b"\0" * (maximum_length - len(content))


def _generate_ec_fragments(
    payload_fragments: list[bytes], *, num_ecc: int
) -> list[bytes]:
    """Construct arrays of error-correction data from a list of QR payloads."""
    logger.debug("Constructing %d error correction QR codes", num_ecc)
    groups = [bytes(x) for x in zip(*payload_fragments)]
    try:
        encoding_rs_codec = reedsolo.RSCodec(num_ecc)
        output_fragments = [
            bytes(x)
            for x in zip(*(encoding_rs_codec.encode(group) for group in groups))
        ]
    except ValueError as e:
        raise QREncodeError("Error: Could not construct error-correction codes.") from e
    return output_fragments


def generate_qr_codes(
    qr_contents: list[bytes], error_tolerance: ECSettingValues = ECSettingValues.M
) -> list[QRCode]:
    """Generate a sequence of QRCodes from ASCII payloads."""
    logger.debug("Generating QR codes")
    image_factory = qr_svg.SvgPathImage
    qr_codes = []
    for code_content in qr_contents:
        code = QRCode(
            version=None,
            error_correction=qr_const.ErrorCorrectionLevel[error_tolerance.name],
            box_size=qr_const.BOX_SIZE,
            image_factory=image_factory,
            border=0,
        )
        logger.debug("Constructing QR with payload: " f"{code_content[:32]!r}...")
        code.add_data(code_content, optimize=0)
        code.make(fit=True)
        logger.debug(f"Code size: {code.version}")
        qr_codes.append(code)
    return qr_codes



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Configuration of the QRDM FastAPI Application."""
import logging
import sys
from enum import Enum
from functools import lru_cache

import structlog
from pydantic_settings import BaseSettings
from structlog.types import EventDict, Processor

from qrdm.models import ECSettingValues

__all__ = ["BackendSettings", "get_backend_settings", "configure_app_logging"]


class LogLevelValues(str, Enum):
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


class BackendSettings(BaseSettings):
    """Pydantic-settings object for QRDM FastAPI Application."""

    QRDM_ERROR_TOLERANCE: ECSettingValues = ECSettingValues.M
    QRDM_ENCODE_EC_CODES: bool = True
    QRDM_LOG_LEVEL: LogLevelValues = LogLevelValues.INFO
    QRDM_JSON_LOGS: bool = False


@lru_cache
def get_backend_settings() -> BackendSettings:
    """Return a cached instance of the application settings."""
    return BackendSettings()


# Adapted from https://gist.github.com/nymous/f138c7f06062b7c43c060bf03759c29e
# MIT License - Copyright (c) 2023 Thomas GAUDIN
def drop_color_message_key(_, __, event_dict: EventDict) -> EventDict:
    """Drop the `color_message` key from an EventDict.

    Uvicorn logs the message a second time in the extra `color_message`, but we don't
    need it. This processor drops the key from the event dict if it exists.
    """
    event_dict.pop("color_message", None)
    return event_dict


def configure_app_logging() -> None:
    """Configure the `structlog` logging pipeline of the FastAPI Application."""
    app_settings = get_backend_settings()

    timestamper = structlog.processors.TimeStamper(fmt="iso")

    shared_processors: list[Processor] = [
        structlog.contextvars.merge_contextvars,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.stdlib.ExtraAdder(),
        drop_color_message_key,
        timestamper,
        structlog.processors.StackInfoRenderer(),
        structlog.processors.UnicodeDecoder(),
    ]

    if app_settings.QRDM_JSON_LOGS:
        # Format the exception only for JSON logs, as we want to pretty-print them when
        # using the ConsoleRenderer
        shared_processors.append(structlog.processors.format_exc_info)

    structlog.configure(
        processors=[
            *shared_processors,
            # Prepare event dict for `ProcessorFormatter`.
            structlog.stdlib.ProcessorFormatter.wrap_for_formatter,
        ],
        logger_factory=structlog.stdlib.LoggerFactory(),
        cache_logger_on_first_use=True,
    )

    log_renderer: structlog.types.Processor
    if app_settings.QRDM_JSON_LOGS:
        log_renderer = structlog.processors.JSONRenderer()
        handler = logging.StreamHandler(sys.stdout)
    else:
        log_renderer = structlog.dev.ConsoleRenderer()
        handler = logging.StreamHandler(sys.stderr)

    formatter = structlog.stdlib.ProcessorFormatter(
        # These run ONLY on `logging` entries that do NOT originate within structlog.
        foreign_pre_chain=shared_processors,
        # These run on ALL entries after the pre_chain is done.
        processors=[
            # Remove _record & _from_structlog.
            structlog.stdlib.ProcessorFormatter.remove_processors_meta,
            log_renderer,
        ],
    )

    # Use OUR `ProcessorFormatter` to format all `logging` entries.
    handler.setFormatter(formatter)
    root_logger = logging.getLogger()
    root_logger.addHandler(handler)
    root_logger.setLevel(app_settings.QRDM_LOG_LEVEL.value)

    for _log in ["uvicorn", "uvicorn.error"]:
        # Clear the log handlers for uvicorn loggers, and enable propagation so the
        # messages are caught by our root logger and formatted correctly by structlog
        logging.getLogger(_log).handlers.clear()
        logging.getLogger(_log).propagate = True

    # Since we re-create the access logs ourselves, to add all information in the
    # structured log (see the `logging_middleware` in main.py), we clear the handlers
    # and prevent the logs to propagate to a logger higher up in the hierarchy
    # (effectively rendering them silent).
    logging.getLogger("uvicorn.access").handlers.clear()
    logging.getLogger("uvicorn.access").propagate = False

    def handle_exception(exc_type, exc_value, exc_traceback):
        """Log any uncaught exception instead of letting it be printed by Python.

        Leaves KeyboardInterrupt untouched to allow users to Ctrl+C to stop
        See https://stackoverflow.com/a/16993115/3641865
        """
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        root_logger.error(
            "Uncaught exception", exc_info=(exc_type, exc_value, exc_traceback)
        )

    sys.excepthook = handle_exception



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""FastAPI Application providing QR encoding and decoding endpoint functions."""

from qrdm.backend.config import configure_app_logging

configure_app_logging()

try:
    from qrdm.backend.main import app
except ImportError as e:
    raise ImportError(
        f"Could not import backend dependency {e.name}. "
        "You can fix this by running `pip install qrdm[backend]`."
    ) from e

__all__ = ["app"]



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT

# This file is just a thin wrapper to enable things like `python setup.py --version`.
# All meaningful setup configuration is contained in pyproject.toml
"""Shim for `setup.py` invocations."""

import setuptools

setuptools.setup()



# Copyright 2024, Massachusetts Institute of Technology
# Subject to FAR 52.227-11 - Patent Rights - Ownership by the Contractor (May 2014).
# SPDX-License-Identifier: MIT
"""Protobuf definitions and classes for serialization of QRDM documents and QR codes."""
from qrdm.qr.protobuf.qrdm_pb2 import DocumentPayload, QRContent

__all__ = ["QRContent", "DocumentPayload", "PROTOBUF_RESERVED_LEN"]

MaxLenQrMeta = QRContent.QRMeta(
    document_hash=2**64 - 1,
    sequence_number=2**32 - 1,
    total_qr_codes=2**32 - 1,
    num_ecc=2**32 - 1,
)

MAX_VARINT_LEN = 10
MAX_QRMETA_LEN = MaxLenQrMeta.ByteSize()
# Each length-delimited field (i.e. the QRMeta submessage and the bytes-type payload)
# field of the message includes a "tag" and "length", each as a varint
PROTOBUF_RESERVED_LEN = MAX_QRMETA_LEN + 4 * MAX_VARINT_LEN
